from pyquest import Room, Interactive
from pyquest import Conditional, Manipulator, Chain, Menu, Break, Teleporter, Random, Lock
from pyquest import Engine
# Defining flags
flags = {
    "player.isSitting": False,
    "inters.tv.on": False,
    "rooms.attic.isLocked": True,
    "inters.trampoline.rand": 0,
    "flags.fbi": False,
    "fillers.fbi": "",
}
# Defining events
events = {
    "inters.couch.getUp": Conditional(
        "player.isSitting",
        {
            "true": Manipulator(
                {"player.isSitting": False},
                "You get up from the couch."
            )
        }
    ),
    "inters.couch.sit": Manipulator(
        {"player.isSitting": True},
        "You sit on the couch."
    ),
    "room.leave": Chain(
        "@inters.couch.getUp",
    ),
    "inters.tv.turnOff": Manipulator({"inters.tv.on": False})
}
# Building rooms
rooms = {
    "outside": Room(
        "You stand outside your house.\n"
        "What are you waiting for? Enter the house."
    ),
    "house.livingRoom": Room(
        "You are at the house living room. \n" 
        "A couch stands in front of a coffee table and a 50 inch television.\n" 
        "You can enter the kitchen, the guest room and the garden.\n" 
        "You can also take the stairs and go up to the second floor, "
        "or leave the house and go out.",
        leave="@room.leave"
    ),
    "house.garden": Room(
        "The garden is breathtaking, with green grass and flowers all around.\n"
        "A huge trampoline is in the center of the garden, waiting to be jumped on.\n"
        "You can enter back into the house."
    ),
    "house.kitchen": Room(
        "The kitchen is accessorized with a fridge, "
        "oven with stove and a sink with some dirty dishes.\n"
        "There is also a dining table you can eat on.\n"
        "You can go out to the living room."
    ),
    "house.guestRoom": Room(
        "This is a usual guest room - with a couch, a bed and a mini bar.\n"
        "When you have guests overnight they stay here.\n"
        "You can go out to the living room.",
        leave="@room.leave"
    ),
    "house.hallway": Room(
        "You are at the hallway of the house second floor.\n"
        "You can enter the bedroom, bathroom and balcony, "
        "or take the stairs and go down the first floor."
    ),
    "house.bedroom": Room(
        "You are at your bedroom.\n"
        "You see a bed, a desk and a painting hanged on the wall.\n"
        "On the desk sits your laptop, a silver, premium Lenovo model.\n"
        "You can go out to the hallway."
    ),
    "house.bathroom": Room(
        "You are at the bathroom.\n"
        "A toilet and a shower stand close to each other.\n"
        "You can go out to the hallway"
    ),
    "house.balcony": Room(
        "You stand on the balcony.\n"
        "A coffee table and two chairs are turning to a nice view of the garden.\n"
        "You can enter back to the hallway."
    ),
    "house.attic": Room(
        "This is the attic. It is a small square space fulled with some interesting stuff.\n"
        "You can go down to the 2nd floor hallway."
    )
}
# Setting up room connections
rooms["outside"].compass.set(
    "in",
    rooms["house.livingRoom"],
    "You unlock the house front door and walk into the living room."
)
rooms["outside"].map.set(
    "house",
    rooms["house.livingRoom"],
    "You unlock the house front door and walk into the living room."
)
rooms["house.livingRoom"].map.set(
    "kitchen",
    rooms["house.kitchen"],
    "You enter the kitchen."
)
rooms["house.livingRoom"].map.set(
    "garden",
    rooms["house.garden"],
    "You open a light glass door which leads to the garden."
)
rooms["house.livingRoom"].map.set(
    "guest room",
    rooms["house.guestRoom"],
    "You open the door and step into the guest room."
)
rooms["house.livingRoom"].compass.set(
    "up",
    rooms["house.hallway"],
    "You climb the grand staircase and reach the hallway of the second floor."
)
rooms["house.livingRoom"].compass.set(
    "out",
    rooms["outside"],
    Chain(
        "You leave the house and close the door behind you.",
        Conditional(
            "flags.fbi",
            {
                "true": Chain(
                    "As soon as you open the front door you see a group armed men in black.\n"
                    "They wear bulletproof FBI suits, and each one of them is holding a gun.\n"
                    "You understand that hacking into the FBI wasn't the best idea ever.",
                    "$game.quit"
                )
            }
        )
    )
)
rooms["house.garden"].map.set(
    "house",
    rooms["house.livingRoom"],
    "You leave the garden and enter the living room."
)
rooms["house.garden"].compass.set(
    "in",
    rooms["house.livingRoom"],
    "You leave the garden and enter the living room."
)
rooms["house.guestRoom"].compass.set(
    "out",
    rooms["house.livingRoom"],
    "You leave the guest room and return to the living room."
)
rooms["house.kitchen"].compass.set(
    "out",
    rooms["house.livingRoom"],
    "You leave the kitchen and return to the living room."
)

rooms["house.hallway"].map.set(
    "bedroom",
    rooms["house.bedroom"],
    "You open your bedroom door and walk in."
)
rooms["house.hallway"].map.set(
    "bathroom",
    rooms["house.bathroom"],
    "You walk into the bathroom."
)
rooms["house.hallway"].map.set(
    "balcony",
    rooms["house.balcony"],
    "You step towards the end of the hallway and go out to the balcony."
)
rooms["house.hallway"].compass.set(
    "out",
    rooms["house.balcony"],
    "You step towards the end of the hallway and go out to the balcony."
)
rooms["house.hallway"].compass.set(
    "down",
    rooms["house.livingRoom"],
    "You climb down the staircase and step into the living room, reaching the ground floor."
)
rooms["house.hallway"].compass.set(
    "up",
    rooms["house.attic"],
    Lock(
        "rooms.attic.isLocked",
        "You climb to the attic through a sliding panel in the celling.",
        "The attic is currently locked!",
    )
)
rooms["house.bedroom"].compass.set(
    "out",
    rooms["house.hallway"],
    "You open the bedroom door and get out to the hallway."
)
rooms["house.bathroom"].compass.set(
    "out",
    rooms["house.hallway"],
    "You open the bathroom door and leave to the hallway."
)
rooms["house.balcony"].compass.set(
    "in",
    rooms["house.hallway"],
    ("You enter back into the house and find yourself in the hallway of the second floor.",
     "The warm sun starts to annoy you and you go back to the second floor hallway.")
)
rooms["house.balcony"].map.set(
    "house",
    rooms["house.hallway"],
    (
        "You enter back into the house and find yourself in the hallway of the second floor.",
        "The warm sun starts to annoy you and you go back to the second floor hallway."
    )
)
rooms["house.balcony"].compass.set(
    "down",
    rooms["house.garden"],
    "You decide to do something stupid and jump of the balcony.\n"
    "You find yourself in the garden, your body hurts like hell.",
    hidden=True
)
rooms["house.attic"].compass.set(
    "down",
    rooms["house.hallway"],
    "You open a panel in the floor and climb down to the 2nd floor hallway."
)
# Making interactives
inters = {
    "couch": Interactive(
        "couch",
        "This is a nice, comfy couch.",
        Conditional(
            "player.isSitting",
            {
                "true": "@inters.couch.getUp",
                "false": "@inters.couch.sit"
            }
        )
    ),
    "tv": Interactive(
        "tv",
        "This is a flat, OLED 4k 50\" television made by Samsung Ltd.",
        Menu(
            Chain(
                Conditional(
                    "inters.tv.on",
                    {
                        "true": "The TV is already on.",
                        "false": Manipulator(
                            {"inters.tv.on": True},
                            "You open the TV."
                        )
                    }
                ),
                "What will you watch?",
            ),
            {
                "Among Us channel": (
                    "You watch a show about the life of an imposter.\n"
                    "The show has a phenomenal deep meaning and it lasts 45 minutes before ending.",
                    "An excited announcer explains the changes and additions in the latest among us update.\n"
                    "Looks like you just wasted 15 minutes of your life.",
                    "'The ultimate guide to not being sus' shows up on screen.\n"
                    "You learn so much about how to not be sus!"
                ),
                "History channel": (
                    "You watch a fascinating documentary film about the rise of the internet.",
                    "You cry to the end of a dramatic movie on the very early beginning of WW2.",
                    "You watch a show about the secret meanings of the bible expressions."
                ),
                "Sports channel": Break(
                    Chain(
                        "You close TV immediately. Watching sports is boring!",
                        "@inters.tv.turnOff"
                    )

                ),
                "Turn TV off": Break("@inters.tv.turnOff")
            },
            True
        )
    ),
    "laptop": Interactive(
        "laptop",
        "This is a premium Lenovo model, with intel core i5 and a nice 15.4 inch display.",
        Chain(
            "You boot up the laptop and get it up and running.",
            Menu(
                "What do you want to do now?",
                {
                    "Go shopping on Ebay": "Ebay is currently down. Sorry!",
                    "Hack into the FBI {fillers.fbi}": Conditional(
                        "flags.fbi",
                        {
                            "false": Manipulator(
                                {"flags.fbi": True, "fillers.fbi": "'(again)'"},
                                "You spend several hours hacking into the FBI.\n"
                                "You do it! You download some data from their database to your computer.\n"
                                "You suddenly hear an aggressive knock on the house front door.\n"
                                "Wonder who it might be?",
                            ),
                            "true": "You hack into the FBI again, but this time you already know how to do it.\n"
                                    "It takes you much shorter period of time."
                        }
                    ),
                    "Turn laptop off": Break("You turn the laptop off.")
                },
                True,
            )
        )
    ),
    "trampoline": Interactive(
        "trampoline",
        "This is a trampoline. You can jump on it.",
        Chain(
            Manipulator({"inters.trampoline.rand": Random(1, 5)}),
            Conditional(
                "inters.trampoline.rand",
                {
                    "5": Teleporter(
                        rooms["house.balcony"],
                        "You jump so hard that you REACH THE BALCONY IN THE SECOND FLOOR!\n"
                        "You launch in the air and fall to the floor of the balcony.\n"
                    ),
                    "!5": "You jump on the trampoline long enough until you are satisfied. "
                }
            )
        )
    ),
    "painting": Interactive(
        "painting",
        "This is a painting of a fantasy dragon and a brave knight in an epic battle.",
        "Your slide the painting to the side, and reveal a small hole in the wall,"
        "approximately the size of a book.\n"
        "This is a safe place to store your killing tools. "
        "No one is supposed to find them here.\n"
    ),
    "bed": Interactive(
        "bed",
        "This is your bed. It is covered with among us bed sheet and blanket.",
        "You enter the bed and close your eyes.\n"
        "You lose your consciousness and find yourself in a dream.\n"
        "Unfortunately, this section of the game hasn't been written yet!"
    ),
    "lever": Interactive(
        "lever",
        "This lever can be used to lock/unlock the door leading to the attic.",
        Conditional(
            "rooms.attic.isLocked",
            {
                "true": Manipulator(
                    {"rooms.attic.isLocked": False},
                    "You pull the lever up and the panel to the attic opens."
                ),
                "false": Manipulator(
                    {"rooms.attic.isLocked": True},
                    "You pull the lever down and the panel to the attic closes."
                ),
            }
        )
    )
}
# Placing interactives
rooms["house.livingRoom"].interactives.add(inters["couch"], inters["tv"])
rooms["house.guestRoom"].interactives.add(inters["couch"])
rooms["house.bedroom"].interactives.add(inters["laptop"], inters["bed"], inters["painting"])
rooms["house.garden"].interactives.add(inters["trampoline"])
rooms["house.hallway"].interactives.add(inters["lever"])


# Adding custom events handler
def handler(event):
    if event == "game.quit":
        print("Game over!")
        exit(0)


def main():
    # Configuring the engine
    engine = Engine(handler, rooms["outside"])
    engine.events = events
    engine.flags = flags
    engine.start()


if __name__ == '__main__':
    main()



import json
import os


FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "defaults.json")


class _Defaults:
    def __init__(self):
        self.data = self._load_defaults()
        self.prompts = self.data["prompts"]
        self.errors = self.data["errors"]
        self.commands = self.data["commands"]

    @staticmethod
    def _load_defaults():
        return json.load(open(FILE, "r", encoding="utf-8"))

    def get(self, query: str) -> str:
        if query.startswith("$"):
            query = query[1:]
        valid_namespaces = ("prompts", "errors", "commands")
        namespace, value = query.split(".", 1)
        if namespace in valid_namespaces:
            return self.data[namespace][value]


defaults = _Defaults()



import random
import re


class Event:
    def __init__(self, event=None, _next=None, **config):
        self.next = _next
        if type(event) == str:
            if event.startswith("$"):
                event = CustomEvent(event[1:])
            elif event.startswith("@"):
                event = NamedEvent(event[1:])
            else:
                event = Prompt(event)
        elif type(event) in (list, tuple):
            event = PromptCarousel(event)
        self.event = event
        self.config = config

    def __str__(self):
        return f"event: {self.event}\nnext: {self.next}"

    def __repr__(self):
        return self.__str__()

class CustomEvent:
    def __init__(self, name):
        self.name = name


class NamedEvent:
    def __init__(self, name):
        self.name = name


class Teleporter:
    def __init__(self, room, event=None):
        self.room = room
        self.event = Event(event)


class Chain:
    def __init__(self, *args):
        args = list(args)
        for i in range(len(args)):
            args[i] = Event(args[i])
        self.events = args


class Conditional:
    def __init__(self, flag_name, options):
        self.flag = flag_name
        self.options = {}
        if type(options) == dict:
            for (name, value) in options.items():
                self.options[name] = Event(value)
        else:
            self.options = {}

    def get(self, flag_value):
        number = re.compile(r"\d+$")
        above = re.compile(r">\d+$")
        below = re.compile(r"<\d+$")
        above_equal = re.compile(r">=\d+$")
        below_equal = re.compile(r"<=\d+$")
        not_equal = False
        string = ""
        # Options: n, <>=n, 'str'
        for (name, value) in self.options.items():
            if name.startswith("!"):
                not_equal = True
                name = name.split("!", 1)[1]
                if number.match(name) is not None:
                    if int(name) != flag_value:
                        return value
            if name.startswith("'") and name.endswith("'") and len(name) > 1:
                string = name[1:-1]
            elif number.match(name) is not None:
                if int(name) == flag_value:
                    return value
            elif name == "true" and flag_value is True:
                return value
            elif name == "false" and flag_value is False:
                return value
            elif name == "null" and flag_value is None:
                return value
            elif above.match(name) is not None:
                if flag_value > int(name[1:]):
                    return value
            elif below.match(name) is not None:
                if flag_value < int(name[1:]):
                    return value
            elif above_equal.match(name) is not None:
                if flag_value >= int(name[2:]):
                    return value
            elif below_equal.match(name) is not None:
                if flag_value <= int(name[2:]):
                    return value

            if string != "":
                if not_equal:
                    if flag_value != name:
                        return value
                else:
                    if flag_value == name:
                        return value
        return Event()


class Manipulator:
    def __init__(self, manipulators, event=None):
        self.event = Event(event)
        self.manipulators = {}
        if type(manipulators) == dict:
            self.manipulators = manipulators

    def manipulate(self, flag, flag_value):
        manipulator = self.manipulators[flag]
        if isinstance(manipulator, Random):
            manipulator = random.randint(manipulator.min, manipulator.max)
        if type(manipulator) != str:
            return manipulator

        number = re.compile(r"\d+$")
        add = re.compile(r"\+\d+$")
        sub = re.compile(r"-\d+$")
        negative = re.compile(r"=-\d+$")
        if manipulator == "null":
            return None
        elif manipulator == "true":
            return True
        elif manipulator == "false":
            return False
        elif manipulator.startswith("'") and manipulator.endswith("'") and len(manipulator) > 1:
            return manipulator[1:-1]
        elif number.match(manipulator) is not None:
            return int(manipulator)
        elif negative.match(manipulator):
            return int(manipulator[1:])
        elif add.match(manipulator) is not None:
            return flag_value + int(manipulator[1:])
        elif sub.match(manipulator) is not None:
            return flag_value - int(manipulator[1:])


class Prompt:
    def __init__(self, prompt):
        self.prompt = prompt


class PromptCarousel:
    def __init__(self, prompts, shuffle=True, repeat=True):
        if type(prompts) in (list, tuple):
            self.prompts = prompts
        else:
            self.prompts = []
        self.current = 0
        self.shuffle = shuffle
        self.repeat = repeat

    def next(self):
        if len(self.prompts) == 0:
            return
        current = self.current
        if current == 0 and self.shuffle:
            # Shuffle the prompts list
            backup = list(self.prompts[:-1])
            index = random.randint(1, len(backup))
            random.shuffle(backup)
            backup.insert(index, self.prompts[-1])
            self.prompts = backup
        self.current += 1
        if self.current == len(self.prompts):
            if self.repeat:
                self.current = 0
            else:
                self.current = current
        return self.prompts[current]


class Break:
    def __init__(self, event=None):
        self.event = Event(event)


class Menu:
    def __init__(self, starter, options, repeat=False):
        self.starter = Event(starter)
        self.repeat = repeat
        self.options = {}
        for (name, value) in options.items():
            self.options[name] = Event(value)


class Random:
    def __init__(self, minimum, maximum=None):
        if maximum is None:
            self.min = 0
            self.max = minimum
        else:
            self.min = minimum
            self.max = maximum


class Lock:
    def __init__(self, flag, success=None, failure=None, final=None):
        self.success = Event(success)
        self.failure = Event(failure)
        self.final = Event(final)
        self.flag = flag



from .events import Chain, NamedEvent, Prompt, PromptCarousel, Menu, Teleporter
from .events import Event, CustomEvent, Conditional, Manipulator, Break, Lock
from .core import Room
from .config import defaults
import re


class Printer:
    def __init__(self):
        pass

    def print(self, *args):
        print(*args)


class Engine:
    def __init__(self, events_handler: callable, starting_room: Room):
        self.currentRoom = starting_room
        self.eventHandler = events_handler
        self.printer = Printer()
        self.run = True

        self.commands = {
            "go": self.command_go,
            "enter": self.command_enter,
            "help": self.command_help,
            "use": self.command_use,
            "activate": self.command_use,
            "look": self.command_look,
            "map": self.command_map,
            "quit": self.quit,
        }
        self.aliases = {
            "exit": "quit",
            "inspect": "look",
            "locations": "map"
        }
        self.flags = {}
        self.events = {}
        self.last_input = ""
        self.prompt = False

    def render(self, namespace, name, *args, output=True):
        valid_namespaces = ("prompts", "errors", "commands")
        if namespace not in valid_namespaces:
            namespace += "s"
        if namespace in valid_namespaces:
            text = defaults.data[namespace][name]
            if len(args) > 0:
                text = text.format(*args)
            if output:
                self.print(text)
            return text

    def print(self, *args: str):
        if len(args) > 0 and type(args[0]) == str and args[0].startswith("$"):
            base = defaults.get(args[0])
            if base is None:
                self.printer.print(*args)
            else:
                if len(args) > 1:
                    args = args[1:]
                    self.printer.print(base.format(*args))
                else:
                    self.printer.print(base)
        else:
            self.printer.print(*args)

    def quit(self, _next=""):
        if _next in ("game", ""):
            self.render("prompt", "game.quit")
            self.run = False
        else:
            self.render("error", "input.invalid")
        
    def custom_event(self, event_name):
        if type(self.eventHandler) == dict:
            if event_name in self.eventHandler:
                if callable(self.eventHandler[event_name]):
                    self.eventHandler[event_name]()
        elif callable(self.eventHandler):
            self.eventHandler(event_name)

    def handle_event(self, wrapper):
        event = wrapper.event
        if isinstance(wrapper.next, Room):
            self.handle_event(self.currentRoom.events["room.leave"])
        if isinstance(event, Chain):
            for e in event.events:
                self.handle_event(e)
        elif isinstance(event, Conditional):
            self.handle_event(event.get(self.flags[event.flag]))
        elif isinstance(event, Manipulator):
            self.handle_event(event.event)
            for flag in event.manipulators:
                self.flags[flag] = event.manipulate(flag, self.flags[flag])
        elif isinstance(event, CustomEvent):
            self.custom_event(event.name)
        elif isinstance(event, NamedEvent):
            if event.name in self.events:
                self.handle_event(Event(self.events[event.name]))
        elif isinstance(event, Prompt):
            prompt = self.parse_prompt(event.prompt)
            self.print(prompt)
        elif isinstance(event, PromptCarousel):
            prompt = self.parse_prompt(event.next())
            self.print(prompt)
        elif isinstance(event, Menu):
            self.show_menu(event)
            self.prompt = True
        elif isinstance(event, Teleporter):
            self.handle_event(event.event)
            self.change_room(event.room, True)
        elif isinstance(event, Lock):
            if self.flags[event.flag] is False:
                self.handle_event(event.success)
                self.handle_event(event.final)
            elif self.flags[event.flag] is True:
                self.handle_event(event.failure)
                self.handle_event(event.final)
                self.prompt = True
                return
        if isinstance(wrapper.next, Room):
            self.change_room(wrapper.next)

    def change_room(self, room, trigger_leave=False):
        if trigger_leave:
            self.handle_event(self.currentRoom.events["room.leave"])
        self.currentRoom = room
        self.handle_event(self.currentRoom.events["room.enter"])
        self.show_prompt()

    def parse_prompt(self, prompt):
        if prompt.count("}") > 0 and prompt.count("{") > 0:
            for (name, value) in self.flags.items():
                filler = "{" + str(name) + "}"
                prompt = prompt.replace(filler, str(value))
        return prompt

    def show_menu(self, menu):
        flag = True
        number = re.compile(r"\d+$")
        choice = 0
        while flag:
            self.handle_event(menu.starter)
            for i in range(len(menu.options)):
                self.print(f"{i+1}. {self.parse_prompt(tuple(menu.options.keys())[i])}")
            while True:
                user_input = input().strip()
                if user_input == "":
                    continue
                if number.match(user_input) is not None:
                    choice = int(user_input)
                    if 0 < choice <= len(menu.options):
                        choice -= 1
                        break
                self.render("error", "menu.choice.invalid")
            event = tuple(menu.options.values())[choice]
            if not menu.repeat:
                flag = False
            if isinstance(event.event, Break):
                flag = False
                event = event.event.event
            self.handle_event(event)
            if flag:
                self.print()

    def command_enter(self, _next):
        found = False
        _map = self.currentRoom.map
        if _next in _map.get_all():
            event = _map.get(_next)
            if event.event is None:
                event.event = Prompt(self.render("prompt", "map.move", _next, output=False))
            self.handle_event(event)

            self.handle_event(_map.get(_next))
            self.prompt = True
            found = True
        if not found:
            self.render("error", "map.notFound", _next)

    def command_go(self, _next):
        compass = self.currentRoom.compass
        result = compass.get(_next)
        if result is None:
            self.render("error", "compass.direction.invalid")
            self.prompt = True
        elif result is False:
            self.render("error", "compass.direction.notFound", _next)
            self.prompt = True
        else:
            event = compass.get(_next)
            if event.event is None:
                event.event = Prompt(self.render("prompt", "compass.move", _next, output=False))
            self.handle_event(event)

    def command_help(self, _next):
        if _next == "":
            self.render("prompts", "game.help")
            for command in defaults.commands:
                self.print("- " + command)
        else:
            found = False
            for command in defaults.commands:
                if _next == command:
                    self.render("prompt", "help.command.description", command.title())
                    self.print(defaults.commands[command])
                    found = True
                    break
            if not found:
                self.render("prompt", "help.command.notFound", _next)
        self.prompt = True

    def command_look(self, _next):
        if _next in ("around", ""):
            self.render("prompt", "room.inspect")
            self.prompt = True
            return
        triggers = ["on the", "at the", "at", "on"]
        for trigger in triggers:
            if _next.startswith(trigger):
                _next = _next.split(trigger, 1)[1].strip()
        interactive = self.currentRoom.interactives.get(_next)
        if interactive is None:
            self.render("error", "interactive.notFound", _next)
        else:
            self.render("prompt", "game.inspect", _next)
            self.handle_event(interactive.description)
            self.prompt = True

    def command_map(self, _next=""):
        room = self.currentRoom
        if room.is_empty():
            self.render("prompt", "room.empty")
        else:
            if len(room.compass) > 0:
                locations = ", ".join(room.compass.get_all(True))
                self.render("prompt", "room.contents.compass", locations)
            if len(room.map) > 0:
                locations = ", ".join(room.map.get_all(True))
                self.render("prompt", "room.contents.map", locations)

            if len(room.interactives) > 0:
                locations = ", ".join(room.interactives.get_all(True))
                self.render("prompt", "room.contents.interactives", locations)
            if len(room.items) > 0:
                locations = ", ".join(room.items.get_all(True))
                self.render("prompt", "room.contents.items", locations)

    def command_use(self, _next):
        room = self.currentRoom
        interactives = room.interactives.get_all()
        if _next in interactives:
            event = room.interactives.get(_next).events["interactive.activate"]
            if isinstance(event, Event):
                self.handle_event(event)
                self.prompt = True
        else:
            self.render("error", "interactive.notFound", _next)

    def show_prompt(self):
        self.prompt = False
        self.handle_event(self.currentRoom.description)
        if self.currentRoom.show_map:
            self.command_map()
        self.handle_input()

    def handle_input(self):
        # Allows quitting the game
        if not self.run:
            return

        # User input loop
        while True:
            # Taking user input
            user_input = input().lower().strip()
            # Building a list of words
            user_input = re.sub(" +", " ", user_input)
            # If there is no input, continue
            if user_input == "":
                continue
            # Special syntax for repeating the last input
            elif user_input == "_":
                user_input = self.last_input
            self.last_input = user_input
            # command = command, the first word
            # _next   = arguments, the following words
            words = user_input.split(" ", 1)
            command = words[0]
            _next = ""
            if len(words) > 1:
                _next = words[1]
            # If input is a valid command, execute that command (help, map, etc.)
            if command in self.commands:
                self.commands[command](_next)
                break
            # Command can also be an alias
            elif command in self.aliases:
                self.commands[self.aliases[command]](_next)
                break
            # Otherwise, show an input error
            else:
                self.render("error", "input.invalid")
        # Prints the current room's prompt if needed
        if self.prompt:
            self.print()
            self.show_prompt()
        # Otherwise, repeats this very function
        else:
            self.handle_input()

    def start(self):
        # Printing the starting prompt
        self.render("prompt", "game.start")
        # Starting mainloop
        self.show_prompt()



from .errors import DirectionError
from .config import defaults
from .events import Event


class _Compass:
    def __init__(self):
        directions = (
            "north",
            "south",
            "east",
            "west",
            "up",
            "down",
            "in",
            "out"
        )
        self.data = {}
        for direction in directions:
            self.data[direction] = None

    def __len__(self):
        return len(self.get_all(prompt=True))

    def set(self, direction, destination, event=None, hidden=False):
        direction = direction.lower().strip()
        for item in self.data:
            if item == direction:
                self.data[item] = Event(event, destination, hidden=hidden)
                return
        raise DirectionError("Undefined direction: " + direction)

    def get(self, direction):
        if len(direction) == 1:
            for item in self.data:
                if item.startswith(direction):
                    return self.data[item]
        else:
            for item in self.data:
                if item == direction:
                    return self.data[direction]
        return False

    def get_all(self, prompt=False):
        result = []
        for (name, value) in self.data.items():
            if value is not None and (not prompt or not value.config["hidden"]):
                result.append(name)
        return result

    def remove(self, direction):
        self.set(direction, None)


class _Map:
    def __init__(self):
        self.data = {}

    def __len__(self):
        return len(self.get_all(prompt=True))

    def set(self, location, destination, event=None, hidden=False):
        if destination is None:
            del self.data[location]
        else:
            self.data[location] = Event(event, destination, hidden=hidden)

    def get(self, location):
        if location in self.data:
            return self.data[location]

    def get_all(self, prompt=False):
        result = []
        for (name, value) in self.data.items():
            if not prompt or not value.config["hidden"]:
                result.append(name)
        return result

    def remove(self, direction):
        self.set(direction, None)


class _Interactives:
    def __init__(self):
        self.data = []

    def __len__(self):
        return len(self.get_all(prompt=True))

    def add(self, *interactives):
        for interactive in interactives:
            added = False
            for i in range(len(self.data)):
                if self.data[i].name == interactive.name:
                    self.data[i] = interactive
                    added = True
            if not added:
                self.data.append(interactive)

    def get(self, name):
        for interactive in self.data:
            if interactive.name == name:
                return interactive

    def get_all(self, prompt=False):
        result = []
        for interactive in self.data:
            if not prompt or not interactive.hidden:
                result.append(interactive.name)
        return result

    def remove(self, name):
        for i in range(len(self.data)):
            if self.data[i].name == name:
                del self.data[i]


class _Items:
    def __init__(self):
        self.data = []

    def __len__(self):
        return len(self.get_all())

    def add(self, item):
        if not isinstance(item, Event):
            self.data.append(Event(item))
        else:
            self.data.append(item)

    def get_all(self):
        result = []
        for item in self.data:
            result.append(item.next.name)
        return result


class Interactive:
    def __init__(self, name, description="", activate=None, inspect=None, hidden=False):
        if description == "":
            description = defaults.prompts["interactive.description"]
        self.name = name
        self.description = Event(description)
        self.hidden = hidden
        self.events = {
            "interactive.inspect": Event(inspect),
            "interactive.activate": Event(activate)
        }

    def set(self, name, value):
        self.events[name] = Event(value)


class Room:
    def __init__(self, description, enter=None, leave=None, show_map=False):
        self.description = Event(description)
        self.compass = _Compass()
        self.map = _Map()
        self.interactives = _Interactives()
        self.items = _Items()
        self.show_map = show_map
        self.events = {
            "room.enter": Event(enter),
            "room.leave": Event(leave),
        }

    def is_empty(self):
        return len(self.map) == 0 and len(self.compass) == 0 and len(self.interactives) == 0 and len(self.items) == 0
